<?php
namespace AristanderAi\Aai\Service;

use AristanderAi\Aai\Block\PageView\Product as ProductBlock;
use AristanderAi\Aai\Block\PageView\ProductFactory;
use AristanderAi\Aai\Model\Event;
use AristanderAi\Aai\Model\EventFactory;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Pricing\SaleableInterface;
use Magento\Catalog\Pricing\Price\FinalPrice;
use Magento\Quote\Model\Quote\Item;
use Magento\Store\Model\Store;
use Magento\Store\Model\StoreManagerInterface;
use Magento\Checkout\Model\Session;

class PageRecorder
{
    /**
     * @var array stores IDs of recorded products to avoid double processing
     */
    protected $recordedProducts;

    /** @var Event */
    protected $event;

    /** @var EventFactory */
    protected $eventFactory;

    /** @var StoreManagerInterface */
    protected $storeManager;

    /** @var Session */
    protected $session;

    /** @var ProductFactory */
    protected $productBlockFactory;

    public function __construct(
        EventFactory $eventFactory,
        StoreManagerInterface $storeManager,
        Session $session,
        ProductFactory $productBlockFactory
    ) {
        $this->eventFactory = $eventFactory;
        $this->storeManager = $storeManager;
        $this->session = $session;
        $this->productBlockFactory = $productBlockFactory;
    }

    /**
     * Starts tracking page view event
     *
     * @return self
     */
    public function start(): self
    {
        $this->event = $this->eventFactory->create(['type' => 'page']);
        $this->event->collect();
        $this->recordedProducts = [];

        return $this;
    }

    /**
     * Finalizes page record data collection and returns event object
     *
     * @return Event|null
     */
    public function exportEvent()
    {
        if (!$this->isStarted()) {
            return null;
        }

        $result = clone $this->event;
        $details = $result->getDetails();

        // Workaround for cart and checkout pages not registering products
        if (isset($details['page_name'])
            && in_array($details['page_name'], ['basket', 'checkout'])
        ) {
            // Collect cart items from quote
            $products = [];
            /** @var Item $item */
            foreach ($this->session->getQuote()->getAllVisibleItems() as $item) {
                $product = $item->getProduct();
                $products[$product->getId()] = $this->extractProductDetails($product);
            }

            $details['products'] = $products;
            $result->setDetails($details);
        }

        return $result;
    }

    /**
     * Tracks product display
     *
     * @param SaleableInterface $product
     * @return string|null
     */
    public function recordProduct(SaleableInterface $product)
    {
        if (in_array($product->getId(), $this->recordedProducts)) {
            return null;
        }

        /** @var ProductBlock $block */
        $block = $this->productBlockFactory->create();
        $block->setDetails($this->extractProductDetails($product));
        $result = $block->toHtml();

        $this->recordedProducts[] = $product->getId();

        return $result;
    }

    /**
     * @param array $products
     * @return self
     */
    public function recordProducts(array $products)
    {
        $details = $this->event->getDetails();
        // Workaround for cart and checkout pages not registering products
        if (empty($products) && isset($details['products'])) {
            $products = $details['products'];
        }

        $details['products'] = array();

        if (isset($details['product_id']) && count($products) > 1) {
            $productId = $details['product_id'];
            assert(isset($products[$productId]), 'Main product not recorded');
            if (isset($products[$productId])) {
                // Put main product to the beginning
                $productItem = $products[$productId];
                unset($products[$productId]);
                array_unshift($products, $productItem);
            }

            unset($details['product_id']);
        }

        // Remove ID keys
        $products = array_values($products);

        $details['products'] = $products;

        $this->event->setDetails($details);

        return $this;
    }

    /**
     * Extracts required details from a given product
     *
     * @param SaleableInterface $product
     * @return array
     */
    public function extractProductDetails(SaleableInterface $product)
    {
        $result = [
            'product_id' => $product->getId(),
        ];

        /** @var FinalPrice $finalPriceModel */
        $finalPriceModel = $product->getPriceInfo()->getPrice(
            FinalPrice::PRICE_CODE);

        /** @var float $min */
        $min = $finalPriceModel->getMinimalPrice()->getValue();
        /** @var float $max */
        $max = $finalPriceModel->getMaximalPrice()->getValue();
        $result['price'] = $min == $max
            ? $min
            : compact('min', 'max');

        /** @var float $min */
        $min = $finalPriceModel->getMinimalPrice()->getAdjustmentAmount('tax');
        if (false !== $min) {
            /** @var float $max */
            $max = $finalPriceModel->getMaximalPrice()->getAdjustmentAmount('tax');
            $result['tax_amount'] = $min == $max
                ? $min
                : compact('min', 'max');
        }

        /** @var Store $store */
        try {
            $store = $this->storeManager->getStore();
            $result['currency_code'] = $store->getCurrentCurrencyCode();
        } catch (NoSuchEntityException $e) {
            // Just do not set currency code
        }

        return $result;
    }

    //
    // Getters and setters
    //

    /**
     * Indicates if page record is enabled and started
     *
     * @return bool
     */
    public function isStarted(): bool
    {
        return $this->event? true : false;
    }

    /**
     * @return Event
     */
    public function getEvent()
    {
        return $this->event;
    }

    /**
     * @param Event $event
     * @return self
     */
    public function setEvent(Event $event)
    {
        $this->event = $event;

        return $this;
    }
}